@@ -45,6 +45,12 @@ _factorial = np.array(
     ]
 )
 
+class Marker(Element):
+
+    _description = []
+
+    def track(self, p):
+        pass
 
 class Drift(Element):
     """Drift in expanded form"""
@@ -74,7 +80,7 @@ class DriftExact(Drift):
         lpzi = length / sqrt(opd ** 2 - p.px ** 2 - p.py ** 2)
         p.x += p.px * lpzi
         p.y += p.py * lpzi
-        p.zeta +=  length - 1 / p.rvv * opd * lpzi
+        p.zeta += length - 1 / p.rvv * opd * lpzi
         p.s += length
 
 
