@@ -8,8 +8,9 @@ from .general import _pkg_root
 from .base_element import BeamElement
 from .beam_elements import *
 from .tracker_data import TrackerData
-from .line import Line
+from .line import Line, Node
 from .tracker import Tracker, freeze_longitudinal
+from .match import Vary, Target
 from .loss_location_refinement import LossLocationRefinement
 from .internal_record import (RecordIdentifier, RecordIndex, new_io_buffer,
                              start_internal_logging, stop_internal_logging)
