@@ -12,10 +12,10 @@ class Table(dict):
         dict.__init__(self, *args, **kwargs)
         self.__dict__ = self
 
-    def to_pandas(self, index=None):
+    def to_pandas(self, index=None,columns=None):
         import pandas as pd
 
-        df = pd.DataFrame(self)
+        df = pd.DataFrame(self,columns=columns)
         if index is not None:
             df.set_index(index, inplace=True)
         return df
