@@ -505,7 +505,9 @@ class MadLoader:
         self.enable_apertures = enable_apertures
         self.classes = classes
         if exact_drift:
-            self._drift = self.classes.DriftExact
+            self._drift = self.classes.DriftExact # will probably be removed
+                                                  # DriftExact is implemented
+                                                  # with compile flag
         else:
             self._drift = self.classes.Drift
         self.ignore_madtypes = ignore_madtypes
@@ -694,7 +696,7 @@ class MadLoader:
         return [self.Builder(mad_elem.name, self._drift, length=mad_elem.l)]
 
     def convert_marker(self, mad_elem):
-        el = self.Builder(mad_elem.name, self._drift, length=0)
+        el = self.Builder(mad_elem.name, self.classes.Marker)
         return self.convert_thin_element([el], mad_elem)
 
     def convert_drift_like(self, mad_elem):
@@ -981,3 +983,25 @@ class MadLoader:
             ee.name, self.classes.FirstOrderTaylorMap, length=length, m0=m0, m1=m1
         )
         return self.convert_thin_element([el], ee)
+
+    def convert_srotation(self, ee):
+        angle = getattr(ee, 'angle')*180/np.pi
+        el = self.Builder(
+            ee.name, self.classes.SRotation, angle=angle
+        )
+        return self.convert_thin_element([el], ee)
+
+    def convert_xrotation(self, ee):
+        angle = getattr(ee, 'angle')*180/np.pi
+        el = self.Builder(
+            ee.name, self.classes.XRotation, angle=angle
+        )
+        return self.convert_thin_element([el], ee)
+
+    def convert_yrotation(self, ee):
+        angle = getattr(ee, 'angle')*180/np.pi
+        el = self.Builder(
+            ee.name, self.classes.YRotation, angle=angle
+        )
+        return self.convert_thin_element([el], ee)
+
