@@ -3,6 +3,8 @@
 # Copyright (c) CERN, 2021.                 #
 # ######################################### #
 
+from time import perf_counter
+from typing import Literal, Union
 import logging
 from functools import partial
 from contextlib import contextmanager
@@ -23,8 +25,10 @@ from .pipeline import PipelineStatus
 from .survey import survey_from_tracker
 from .tracker_data import TrackerData
 from .twiss import (compute_one_turn_matrix_finite_differences,
-                    find_closed_orbit, match_tracker, twiss_from_tracker)
+                    find_closed_orbit, twiss_from_tracker)
+from .match import match_tracker
 from .tapering import compensate_radiation_energy_loss
+from .prebuild_kernels import get_suitable_kernel, XT_PREBUILT_KERNELS_LOCATION
 
 logger = logging.getLogger(__name__)
 
@@ -33,8 +37,15 @@ def _check_is_collective(ele):
     iscoll = not hasattr(ele, 'iscollective') or ele.iscollective
     return iscoll
 
+
 class Tracker:
 
+    '''
+    Xsuite tracker class. It is the core of the xsuite package, allows tracking
+    particles in a given beam line. Methods to match particle distributions
+    and to compute twiss parameters are also available.
+    '''
+
     def __init__(
         self,
         _context=None,
@@ -51,14 +62,15 @@ class Tracker:
         global_xy_limit=1.0,
         extra_headers=(),
         local_particle_src=None,
-        save_source_as=None,
         io_buffer=None,
         compile=True,
+        use_prebuilt_kernels=True,
         enable_pipeline_hold=False,
         _element_ref_data=None,
     ):
         self.config = TrackerConfig()
-        self.config.XTRACK_MULTIPOLE_NO_SYNRAD=True
+        self.config.XTRACK_MULTIPOLE_NO_SYNRAD = True
+        self.config.XFIELDS_BB3D_NO_BEAMSTR = True
 
         if sequence is not None:
             raise ValueError(
@@ -93,9 +105,9 @@ class Tracker:
                 global_xy_limit=global_xy_limit,
                 extra_headers=extra_headers,
                 local_particle_src=local_particle_src,
-                save_source_as=save_source_as,
                 io_buffer=io_buffer,
                 compile=compile,
+                use_prebuilt_kernels=use_prebuilt_kernels,
                 enable_pipeline_hold=enable_pipeline_hold)
         else:
             self._element_ref_data = _element_ref_data
@@ -113,9 +125,9 @@ class Tracker:
                 global_xy_limit=global_xy_limit,
                 extra_headers=extra_headers,
                 local_particle_src=local_particle_src,
-                save_source_as=save_source_as,
                 io_buffer=io_buffer,
                 compile=compile,
+                use_prebuilt_kernels=use_prebuilt_kernels,
                 enable_pipeline_hold=enable_pipeline_hold)
 
         self.matrix_responsiveness_tol = lnf.DEFAULT_MATRIX_RESPONSIVENESS_TOL
@@ -136,9 +148,9 @@ class Tracker:
         global_xy_limit=1.0,
         extra_headers=(),
         local_particle_src=None,
-        save_source_as=None,
         io_buffer=None,
         compile=True,
+        use_prebuilt_kernels=True,
         enable_pipeline_hold=False
     ):
 
@@ -153,8 +165,8 @@ class Tracker:
         self.global_xy_limit = global_xy_limit
         self.extra_headers = extra_headers
         self.local_particle_src = local_particle_src
-        self.save_source_as = save_source_as
         self._enable_pipeline_hold = enable_pipeline_hold
+        self.use_prebuilt_kernels = use_prebuilt_kernels
 
         if _buffer is None:
             if _context is None:
@@ -233,8 +245,8 @@ class Tracker:
                 extra_headers=extra_headers,
                 reset_s_at_end_turn=reset_s_at_end_turn,
                 local_particle_src=local_particle_src,
-                save_source_as=save_source_as,
-                io_buffer=self.io_buffer
+                io_buffer=self.io_buffer,
+                use_prebuilt_kernels=use_prebuilt_kernels,
                 )
         supertracker.config = self.config
 
@@ -251,7 +263,8 @@ class Tracker:
                                 extra_headers=extra_headers,
                                 local_particle_src=local_particle_src,
                                 skip_end_turn_actions=True,
-                                io_buffer=self.io_buffer)
+                                io_buffer=self.io_buffer,
+                                use_prebuilt_kernels=use_prebuilt_kernels,)
                 parts[ii].config = self.config
 
         # Make a "marker" element to increase at_element
@@ -273,6 +286,7 @@ class Tracker:
         self._element_part = _element_part
         self._element_index_in_part = _element_index_in_part
         self._radiation_model = None
+        self._beamstrahlung_model = None
 
     def _init_track_no_collective(
         self,
@@ -289,9 +303,9 @@ class Tracker:
         global_xy_limit=1.0,
         extra_headers=(),
         local_particle_src=None,
-        save_source_as=None,
         io_buffer=None,
         compile=True,
+        use_prebuilt_kernels=True,
         enable_pipeline_hold=False
     ):
         if track_kernel == 'skip':
@@ -345,24 +359,86 @@ class Tracker:
         self.line = line
         self.line.tracker = self
         self._tracker_data = tracker_data
+        self.num_elements = len(tracker_data.elements)
+        self._buffer = tracker_data._buffer
 
         self.particles_class = particles_class
         self.particles_monitor_class = particles_monitor_class
-        self.num_elements = len(tracker_data.elements)
         self.global_xy_limit = global_xy_limit
         self.extra_headers = extra_headers
         self.skip_end_turn_actions = skip_end_turn_actions
         self.reset_s_at_end_turn = reset_s_at_end_turn
         self.local_particle_src = local_particle_src
         self.element_classes = element_classes
-        self._buffer = tracker_data._buffer
 
-        if track_kernel is None:
-            track_kernel = {}
-        self.track_kernel = track_kernel
+        self.track_kernel = track_kernel or {}
 
         self.track = self._track_no_collective
         self._radiation_model = None
+        self._beamstrahlung_model = None
+        self.use_prebuilt_kernels = use_prebuilt_kernels
+
+        if compile:
+            _ = self._current_track_kernel  # This triggers compilation
+
+    def optimize_for_tracking(self, compile=True, verbose=True, keep_markers=False):
+        """
+        Optimize the tracker for tracking speed.
+        """
+        if self.iscollective:
+            raise NotImplementedError("Optimization is not implemented for "
+                                      "collective trackers")
+
+        self.track_kernel = {} # Remove all kernels
+
+        if verbose: print("Disable xdeps expressions")
+        self.line._var_management = None # Disable expressions
+
+        line = self.line
+
+        # Unfreeze the line
+        line.element_names = list(line.element_names)
+
+        if keep_markers is True:
+            if verbose: print('Markers are kept')
+        elif keep_markers is False:
+            if verbose: print("Remove markers")
+            line.remove_markers()
+        else:
+            if verbose: print('Keeping only selected markers')
+            line.remove_markers(keep=keep_markers)
+
+        if verbose: print("Remove inactive multipoles")
+        line.remove_inactive_multipoles()
+
+        if verbose: print("Merge consecutive multipoles")
+        line.merge_consecutive_multipoles()
+
+        if verbose: print("Remove zero length drifts")
+        line.remove_zero_length_drifts()
+
+        if verbose: print("Merge consecutive drifts")
+        line.merge_consecutive_drifts()
+
+        if verbose: print("Use simple bends")
+        line.use_simple_bends()
+
+        if verbose: print("Use simple quadrupoles")
+        line.use_simple_quadrupoles()
+
+        if verbose: print("Rebuild tracker data")
+        tracker_data = TrackerData(
+            line=line,
+            extra_element_classes=(self.particles_monitor_class._XoStruct,),
+            _buffer=self._buffer)
+
+        self.line._freeze()
+
+        self._tracker_data = tracker_data
+        self.element_classes = tracker_data.element_classes
+        self.num_elements = len(tracker_data.elements)
+
+        self.use_prebuilt_kernels = False
 
         if compile:
             _ = self._current_track_kernel # This triggers compilation
@@ -383,7 +459,8 @@ class Tracker:
                 "Please rebuild the tracker, for example using `line.build_tracker(...)`.")
 
     def find_closed_orbit(self, particle_co_guess=None, particle_ref=None,
-                          co_search_settings={}, delta_zeta=0, delta0=None,
+                          co_search_settings={}, delta_zeta=0,
+                          delta0=None, zeta0=None,
                           continue_on_closed_orbit_error=False,
                           freeze_longitudinal=False):
 
@@ -409,7 +486,7 @@ class Tracker:
             tracker = self
 
         return find_closed_orbit(tracker, particle_co_guess=particle_co_guess,
-                                 particle_ref=particle_ref, delta0=delta0,
+                                 particle_ref=particle_ref, delta0=delta0, zeta0=zeta0,
                                  co_search_settings=co_search_settings, delta_zeta=delta_zeta,
                                  continue_on_closed_orbit_error=continue_on_closed_orbit_error)
 
@@ -430,7 +507,7 @@ class Tracker:
         return compute_one_turn_matrix_finite_differences(tracker, particle_on_co,
                                                    steps_r_matrix)
 
-    def twiss(self, particle_ref=None, delta0=None, method='6d',
+    def twiss(self, particle_ref=None, delta0=None, zeta0=None, method='6d',
         r_sigma=0.01, nemitt_x=1e-6, nemitt_y=1e-6,
         delta_disp=1e-5, delta_chrom=1e-4,
         particle_co_guess=None, R_matrix=None, W_matrix=None,
@@ -462,10 +539,18 @@ class Tracker:
                                    reverse=reverse)
 
     def match(self, vary, targets, **kwargs):
+        '''
+        Change a set of knobs in the beamline in order to match assigned targets.
+        See corresponding section is the Xsuite User's guide.
+        '''
         return match_tracker(self, vary, targets, **kwargs)
 
     def filter_elements(self, mask=None, exclude_types_starting_with=None):
 
+        """
+        Replace with Drifts all elements satisfying a given condition.
+        """
+
         self._check_invalidated()
 
         return self.__class__(
@@ -477,14 +562,23 @@ class Tracker:
                  element_classes=(self.element_classes if not self.iscollective
                                     else self._supertracker.element_classes))
 
-    def configure_radiation(self, model=None, mode=None):
+    def configure_radiation(self, model=None, model_beamstrahlung=None,
+                            mode='deprecated'):
 
-        if mode is not None:
+        """
+        Configure synchrotron radiation and beamstrahlung models.
+        Choose among: None / "mean"/ "quantum".
+        See corresponding section is the Xsuite User's guide.
+        """
+
+        if mode != 'deprecated':
             raise NameError('mode is deprecated, use model instead')
 
         self._check_invalidated()
 
         assert model in [None, 'mean', 'quantum']
+        assert model_beamstrahlung in [None, 'mean', 'quantum']
+
         if model == 'mean':
             radiation_flag = 1
             self._radiation_model = 'mean'
@@ -495,18 +589,39 @@ class Tracker:
             radiation_flag = 0
             self._radiation_model = None
 
+        if model_beamstrahlung == 'mean':
+            beamstrahlung_flag = 1
+            self._beamstrahlung_model = 'mean'
+        elif model_beamstrahlung == 'quantum':
+            beamstrahlung_flag = 2
+            self._beamstrahlung_model = 'quantum'
+        else:
+            beamstrahlung_flag = 0
+            self._beamstrahlung_model = None
+
         for kk, ee in self.line.element_dict.items():
             if hasattr(ee, 'radiation_flag'):
                 ee.radiation_flag = radiation_flag
 
-        if radiation_flag == 2:
+        for kk, ee in self.line.element_dict.items():
+            if hasattr(ee, 'flag_beamstrahlung'):
+                ee.flag_beamstrahlung = beamstrahlung_flag
+
+        if radiation_flag == 2 or beamstrahlung_flag == 2:
             self.line._needs_rng = True
 
-        self.config.XTRACK_MULTIPOLE_NO_SYNRAD = False
+        self.config.XTRACK_MULTIPOLE_NO_SYNRAD = (radiation_flag == 0)
+        self.config.XFIELDS_BB3D_NO_BEAMSTR = (beamstrahlung_flag == 0)
 
     def compensate_radiation_energy_loss(self, delta0=0, rtot_eneloss=1e-10,
                                     max_iter=100, **kwargs):
 
+        """
+        Compensate beam energy loss from synchrotron radiation by configuring
+        RF cavities and Multipole elements (tapering).
+        See corresponding section is the Xsuite User's guide.
+        """
+
         all_kwargs = locals().copy()
         all_kwargs.pop('self')
         all_kwargs.pop('kwargs')
@@ -516,6 +631,10 @@ class Tracker:
     def cycle(self, index_first_element=None, name_first_element=None,
               _buffer=None, _context=None):
 
+        """
+        Cycle the line to start from a given element.
+        """
+
         self._check_invalidated()
 
         cline = self.line.cycle(index_first_element=index_first_element,
@@ -543,12 +662,21 @@ class Tracker:
             )
 
     def build_particles(self, *args, **kwargs):
+
+        """
+        Generate a particle distribution. Equivalent to xp.Particles(tracker=tracker, ...)
+        See corresponding section is the Xsuite User's guide.
+        """
         res = xp.build_particles(*args, tracker=self, **kwargs)
         return res
 
     def get_backtracker(self, _context=None, _buffer=None,
                         global_xy_limit='from_tracker'):
 
+        """
+        Build a Tracker object that backtracks in the same line.
+        """
+
         self._check_invalidated()
 
         assert not self.iscollective
@@ -620,11 +748,38 @@ class Tracker:
     def _context(self):
         return self._buffer.context
 
-    def _build_kernel(self, save_source_as, compile):
+    def _build_kernel(
+            self,
+            compile: Union[bool, Literal['force']],
+            module_name=None,
+            containing_dir='.',
+    ):
+        if (self.use_prebuilt_kernels and compile != 'force'
+                and isinstance(self._context, xo.ContextCpu)):
+            kernel_info = get_suitable_kernel(
+                self.config, self.element_classes
+            )
+            if kernel_info:
+                module_name, modules_classes = kernel_info
+                kernel_description = self.get_kernel_descriptions()['track_line']
+                kernels = self._context.kernels_from_file(
+                    module_name=module_name,
+                    containing_dir=XT_PREBUILT_KERNELS_LOCATION,
+                    kernel_descriptions={'track_line': kernel_description},
+                )
+                self._context.kernels.update(kernels)
+                self._current_track_kernel = self._context.kernels['track_line']
+                self.element_classes = [cls._XoStruct for cls in modules_classes]
+                self._tracker_data = TrackerData(
+                    line=self.line,
+                    element_classes=self.element_classes,
+                    _context=self._context,
+                    _buffer=self._buffer,
+                )
+                return
 
         context = self._tracker_data._buffer.context
 
-        kernels = {}
         headers = []
 
         headers.extend(self.extra_headers)
@@ -693,7 +848,7 @@ class Tracker:
 
                         // Get the pointer to and the type id of the `elem_idx`th
                         // element in `element_ref_data.elements`:
-                        void* el = ElementRefData_member_elements(elem_ref_data, elem_idx);
+                        /*gpuglmem*/ void* el = ElementRefData_member_elements(elem_ref_data, elem_idx);
                         int64_t elem_type = ElementRefData_typeid_elements(elem_ref_data, elem_idx);
 
                         switch(elem_type){
@@ -756,8 +911,42 @@ class Tracker:
 
         source_track = "\n".join(src_lines)
 
+        kernels = self.get_kernel_descriptions(context)
+
+        # Compile!
+        if isinstance(self._context, xo.ContextCpu):
+            kwargs = {
+                'containing_dir': containing_dir,
+                'module_name': module_name,
+            }
+        else:
+            # Saving kernels is unsupported on GPU
+            kwargs = {}
+
+        out_kernels = context.build_kernels(
+            sources=[source_track],
+            kernel_descriptions=kernels,
+            extra_headers=self._config_to_headers() + headers,
+            extra_classes=self.element_classes,
+            apply_to_source=[
+                partial(_handle_per_particle_blocks,
+                        local_particle_src=self.local_particle_src)],
+            specialize=True,
+            compile=compile,
+            save_source_as=f'{module_name}.c' if module_name else None,
+            **kwargs,
+        )
+        context.kernels.update(out_kernels)
+
+        self._current_track_kernel = context.kernels.track_line
+
+    def get_kernel_descriptions(self, _context=None):
+        if not _context:
+            _context = self._context
+
         kernel_descriptions = {
             "track_line": xo.Kernel(
+                c_name='track_line',
                 args=[
                     xo.Arg(xo.Int8, pointer=True, name="buffer"),
                     xo.Arg(self._tracker_data._element_ref_data.__class__, name="tracker_data"),
@@ -775,29 +964,10 @@ class Tracker:
             )
         }
 
-        # Internal API can be exposed only on CPU
-        if not isinstance(context, xo.ContextCpu):
-            kernels = {}
-        kernels.update(kernel_descriptions)
-
         # Random number generator init kernel
-        kernels.update(self.particles_class._kernels)
+        kernel_descriptions.update(self.particles_class._kernels)
 
-        # Compile!
-        context.add_kernels(
-            [source_track],
-            kernels,
-            extra_headers=self._config_to_headers() + headers,
-            extra_classes=self.element_classes,
-            apply_to_source=[
-                partial(_handle_per_particle_blocks,
-                        local_particle_src=self.local_particle_src)],
-            save_source_as=save_source_as,
-            specialize=True,
-            compile=compile
-        )
-
-        self._current_track_kernel = context.kernels.track_line
+        return kernel_descriptions
 
     def _prepare_collective_track_session(self, particles, ele_start, ele_stop,
                                        num_elements, num_turns, turn_by_turn_monitor):
@@ -950,17 +1120,26 @@ class Tracker:
         return stop_tracking, skip, ret
 
     def resume(self, session):
+        """
+        Resume a track session that had been placed on hold.
+        """
         return self._track_with_collective(particles=None, _session_to_resume=session)
 
     def freeze_vars(self, variable_names):
+        """Freeze assigned coordinates in tracked Particles objects."""
         for name in variable_names:
             self.config[f'FREEZE_VAR_{name}'] = True
 
     def unfreeze_vars(self, variable_names):
+        """Unfreeze variables previously frozen with `freeze_vars`."""
         for name in variable_names:
             self.config[f'FREEZE_VAR_{name}'] = False
 
     def freeze_longitudinal(self, state=True):
+        """
+        Freeze longitudinal coordinates in tracked Particles objects.
+        See corresponding section is the Xsuite User's guide.
+        """
         assert state in (True, False)
         assert self.iscollective is False, ('Cannot freeze longitudinal '
                         'variables in collective mode (not yet implemented)')
@@ -978,8 +1157,13 @@ class Tracker:
         num_turns=None,    # defaults to 1
         turn_by_turn_monitor=None,
         freeze_longitudinal=False,
+        time=False,
         _session_to_resume=None
     ):
+
+        if time:
+            t0 = perf_counter()
+
         if freeze_longitudinal:
             raise NotImplementedError('freeze_longitudinal not implemented yet'
                                       ' for collective tracking')
@@ -1130,6 +1314,13 @@ class Tracker:
 
         self.record_last_track = monitor
 
+        if time:
+            t1 = perf_counter()
+            self._context.synchronize()
+            self.time_last_track = t1 - t0
+        else:
+            self.time_last_track = None
+
     def _track_no_collective(
         self,
         particles,
@@ -1138,9 +1329,13 @@ class Tracker:
         num_elements=None, # defaults to full lattice
         num_turns=None,    # defaults to 1
         turn_by_turn_monitor=None,
-        freeze_longitudinal=False
+        freeze_longitudinal=False,
+        time=False
     ):
 
+        if time:
+            t0 = perf_counter()
+
         if freeze_longitudinal:
             kwargs = locals().copy()
             kwargs.pop('self')
@@ -1312,6 +1507,13 @@ class Tracker:
 
         self.record_last_track = monitor
 
+        if time:
+            self._context.synchronize()
+            t1 = perf_counter()
+            self.time_last_track = t1 - t0
+        else:
+            self.time_last_track = None
+
     @staticmethod
     def _get_default_monitor_class():
         import xtrack as xt  # I have to do it like this
@@ -1355,10 +1557,18 @@ class Tracker:
 
     def start_internal_logging_for_elements_of_type(self,
                                                     element_type, capacity):
+        """
+        Start internal logging for all elements of a given type.
+        See corresponding section is the Xsuite User's guide.
+        """
         return start_internal_logging_for_elements_of_type(self,
                                                     element_type, capacity)
 
     def stop_internal_logging_for_elements_of_type(self, element_type):
+        """
+        Stop internal logging for all elements of a given type.
+        See corresponding section is the Xsuite User's guide.
+        """
         self._check_invalidated()
         stop_internal_logging_for_elements_of_type(self, element_type)
 
@@ -1447,7 +1657,7 @@ class Tracker:
         try:
             return self.track_kernel[self._hashable_config()]
         except KeyError:
-            self._build_kernel(save_source_as=None, compile=True)
+            self._build_kernel(compile=True)
             return self._current_track_kernel
 
     @_current_track_kernel.setter
@@ -1473,7 +1683,10 @@ def freeze_longitudinal(tracker):
         yield None
     finally:
         tracker.config = config
-_freeze_longitudinal = freeze_longitudinal # to avoid name clash with function argument
+
+
+_freeze_longitudinal = freeze_longitudinal  # to avoid name clash with function argument
+
 
 class TrackerConfig(dict):
     def __init__(self, *args, **kwargs):
@@ -1487,4 +1700,13 @@ class TrackerConfig(dict):
             super(TrackerConfig, self).__setitem__(idx, val)
 
     def __setattr__(self, idx, val):
-        self[idx] = val
+        if val is not False:
+            self[idx] = val
+        elif idx in self:
+            del(self[idx])
+
+    def update(self, other, **kwargs):
+        super().update(other, **kwargs)
+        keys_for_none_vals = [k for k, v in self.items() if v is False]
+        for k in keys_for_none_vals:
+            del self[k]
