@@ -5,19 +5,20 @@
 
 import logging
 import warnings
-from functools import partial
-from operator import ne
+
 import numpy as np
+from scipy.optimize import fsolve
 
 import xobjects as xo
 import xpart as xp
 
-from scipy.optimize import fsolve
+
 from scipy.constants import c as clight
 
 from . import linear_normal_form as lnf
 from .general import Table
 
+
 import xtrack as xt # To avoid circular imports
 
 DEFAULT_STEPS_R_MATRIX = {
@@ -28,10 +29,13 @@ DEFAULT_STEPS_R_MATRIX = {
 
 DEFAULT_CO_SEARCH_TOL = [1e-12, 1e-12, 1e-12, 1e-12, 1e-5, 1e-12]
 
+AT_TURN_FOR_TWISS = -10 # # To avoid writing in monitors installed in the line
+
 log = logging.getLogger(__name__)
 
 def twiss_from_tracker(tracker, particle_ref=None, method='6d',
-        particle_on_co=None, R_matrix=None, W_matrix=None, delta0=None,
+        particle_on_co=None, R_matrix=None, W_matrix=None,
+        delta0=None, zeta0=None,
         r_sigma=0.01, nemitt_x=1e-6, nemitt_y=2.5e-6,
         delta_disp=1e-5, delta_chrom = 1e-4,
         particle_co_guess=None, steps_r_matrix=None,
@@ -111,17 +115,20 @@ def twiss_from_tracker(tracker, particle_ref=None, method='6d',
         assert at_elements is None
         (auxtracker, names_inserted_markers
             ) = _build_auxiliary_tracker_with_extra_markers(
-            tracker=tracker, at_s=at_s, marker_prefix='inserted_twiss_marker')
+            tracker=tracker, at_s=at_s, marker_prefix='inserted_twiss_marker',
+            algorithm='insert')
         kwargs.pop('tracker')
         kwargs.pop('at_s')
         kwargs.pop('at_elements')
         kwargs.pop('matrix_responsiveness_tol')
         kwargs.pop('matrix_stability_tol')
-        return twiss_from_tracker(tracker=auxtracker,
+        res = twiss_from_tracker(tracker=auxtracker,
                         at_elements=names_inserted_markers,
                         matrix_responsiveness_tol=matrix_responsiveness_tol,
                         matrix_stability_tol=matrix_stability_tol,
                         **kwargs)
+        #res._auxtracker = auxtracker # DEBUG
+        return res
 
     mux0 = 0
     muy0 = 0
@@ -161,7 +168,8 @@ def twiss_from_tracker(tracker, particle_ref=None, method='6d',
                                 particle_ref=particle_ref,
                                 co_search_settings=co_search_settings,
                                 continue_on_closed_orbit_error=continue_on_closed_orbit_error,
-                                delta0=delta0)
+                                delta0=delta0,
+                                zeta0=zeta0)
 
     if W_matrix is not None:
         W = W_matrix
@@ -186,12 +194,14 @@ def twiss_from_tracker(tracker, particle_ref=None, method='6d',
                             particle_ref=particle_ref,
                             co_search_settings=co_search_settings,
                             continue_on_closed_orbit_error=continue_on_closed_orbit_error,
-                            delta0=delta0-delta_disp)
+                            delta0=delta0-delta_disp,
+                            zeta0=zeta0)
         p_disp_plus = tracker.find_closed_orbit(particle_co_guess=particle_co_guess,
                             particle_ref=particle_ref,
                             co_search_settings=co_search_settings,
                             continue_on_closed_orbit_error=continue_on_closed_orbit_error,
-                            delta0=delta0+delta_disp)
+                            delta0=delta0+delta_disp,
+                            zeta0=zeta0)
         p_disp_minus.move(_context=xo.context_default)
         p_disp_plus.move(_context=xo.context_default)
         dx_dpzeta = ((p_disp_plus.x[0] - p_disp_minus.x[0])
@@ -232,6 +242,8 @@ def twiss_from_tracker(tracker, particle_ref=None, method='6d',
     twiss_res['circumference'] = circumference
 
     if not skip_global_quantities:
+
+        s_vect = twiss_res_element_by_element['s']
         mux = twiss_res_element_by_element['mux']
         muy = twiss_res_element_by_element['muy']
 
@@ -256,11 +268,25 @@ def twiss_from_tracker(tracker, particle_ref=None, method='6d',
         betz0 = W[4, 4]**2 + W[4, 5]**2
         ptau_co = twiss_res_element_by_element['ptau']
 
+        # Coupling
+        r1 = (np.sqrt(twiss_res_element_by_element['bety1'])/
+              np.sqrt(twiss_res_element_by_element['betx1']))
+        r2 = (np.sqrt(twiss_res_element_by_element['betx2'])/
+              np.sqrt(twiss_res_element_by_element['bety2']))
+
+        # Coupling (https://arxiv.org/pdf/2005.02753.pdf)
+        cmin_arr = (2 * np.sqrt(r1*r2) *
+                    np.abs(np.mod(mux[-1], 1) - np.mod(muy[-1], 1))
+                    /(1 + r1 * r2))
+        c_minus = np.trapz(cmin_arr, s_vect)/(circumference)
+        c_r1_avg = np.trapz(r1, s_vect)/(circumference)
+        c_r2_avg = np.trapz(r2, s_vect)/(circumference)
         twiss_res.update({
             'qx': mux[-1], 'qy': muy[-1], 'qs': qs, 'dqx': dqx, 'dqy': dqy,
             'slip_factor': eta, 'momentum_compaction_factor': alpha, 'betz0': betz0,
             'circumference': circumference, 'T_rev': T_rev,
-            'particle_on_co':part_on_co.copy(_context=xo.context_default)
+            'particle_on_co':part_on_co.copy(_context=xo.context_default),
+            'c_minus': c_minus, 'c_r1_avg': c_r1_avg, 'c_r2_avg': c_r2_avg
         })
         if hasattr(part_on_co, '_fsolve_info'):
             twiss_res['particle_on_co']._fsolve_info = part_on_co._fsolve_info
@@ -335,13 +361,16 @@ def _propagate_optics(tracker, W_matrix, particle_on_co,
     part_for_twiss.at_element = particle_on_co._xobject.at_element[0]
     i_start = part_for_twiss._xobject.at_element[0]
 
-    assert np.all(ctx2np(part_for_twiss.at_turn) == 0)
+    part_for_twiss.at_turn = AT_TURN_FOR_TWISS # To avoid writing in monitors
+
+    #assert np.all(ctx2np(part_for_twiss.at_turn) == 0)
     tracker.track(part_for_twiss, turn_by_turn_monitor='ONE_TURN_EBE',
                   ele_start=ele_start, ele_stop=ele_stop)
     assert np.all(ctx2np(part_for_twiss.state) == 1), (
         'Some test particles were lost during twiss!')
     i_stop = part_for_twiss._xobject.at_element[0] + (
-        part_for_twiss._xobject.at_turn[0] * len(tracker.line.element_names))
+        (part_for_twiss._xobject.at_turn[0] - AT_TURN_FOR_TWISS)
+         * len(tracker.line.element_names))
 
     x_co = tracker.record_last_track.x[6, i_start:i_stop+1].copy()
     y_co = tracker.record_last_track.y[6, i_start:i_stop+1].copy()
@@ -405,7 +434,7 @@ def _propagate_optics(tracker, W_matrix, particle_on_co,
     # Computation of twiss parameters
 
     if use_full_inverse:
-        betx, alfx, gamx, bety, alfy, gamy = _extract_twiss_parameters_with_inverse(Ws)
+        betx, alfx, gamx, bety, alfy, gamy, bety1, betx2 = _extract_twiss_parameters_with_inverse(Ws)
     else:
         betx = Ws[:, 0, 0]**2 + Ws[:, 0, 1]**2
         bety = Ws[:, 2, 2]**2 + Ws[:, 2, 3]**2
@@ -416,6 +445,12 @@ def _propagate_optics(tracker, W_matrix, particle_on_co,
         alfx = - Ws[:, 0, 0] * Ws[:, 1, 0] - Ws[:, 0, 1] * Ws[:, 1, 1]
         alfy = - Ws[:, 2, 2] * Ws[:, 3, 2] - Ws[:, 2, 3] * Ws[:, 3, 3]
 
+        bety1 = Ws[:, 2, 0]**2 + Ws[:, 2, 1]**2
+        betx2 = Ws[:, 0, 2]**2 + Ws[:, 0, 3]**2
+
+    betx1 = betx
+    bety2 = bety
+
     mux = np.unwrap(phix)/2/np.pi
     muy = np.unwrap(phiy)/2/np.pi
     muzeta = np.unwrap(phizeta)/2/np.pi
@@ -454,6 +489,10 @@ def _propagate_optics(tracker, W_matrix, particle_on_co,
         'muy': muy,
         'muzeta': muzeta,
         'W_matrix': W_matrix,
+        'betx1': betx1,
+        'bety1': bety1,
+        'betx2': betx2,
+        'bety2': bety2,
     }
 
     return twiss_res_element_by_element
@@ -559,7 +598,8 @@ class ClosedOrbitSearchError(Exception):
     pass
 
 def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
-                      co_search_settings=None, delta_zeta=0, delta0=None,
+                      co_search_settings=None, delta_zeta=0,
+                      delta0=None, zeta0=None,
                       continue_on_closed_orbit_error=False):
 
     if particle_co_guess is None:
@@ -603,12 +643,19 @@ def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
                     particle_co_guess._xobject.py[0] + shift_factor * 1e-7,
                     particle_co_guess._xobject.zeta[0] + shift_factor * 1e-4,
                     particle_co_guess._xobject.delta[0] + shift_factor * 1e-5])
-        if delta0 is not None:
+        if delta0 is not None and zeta0 is None:
             x0[5] = delta0
-            _error_for_co = _error_for_co_search_4d
+            _error_for_co = _error_for_co_search_4d_delta0
+        elif delta0 is None and zeta0 is not None:
+            x0[4] = zeta0
+            _error_for_co = _error_for_co_search_4d_zeta0
+        elif delta0 is not None and zeta0 is not None:
+            _error_for_co = _error_for_co_search_4d_delta0_zeta0
         else:
             _error_for_co = _error_for_co_search_6d
-        if np.all(np.abs(_error_for_co(x0, particle_co_guess, tracker, delta_zeta, delta0))
+        if zeta0 is not None:
+            x0[4] = zeta0
+        if np.all(np.abs(_error_for_co(x0, particle_co_guess, tracker, delta_zeta, delta0, zeta0))
                     < DEFAULT_CO_SEARCH_TOL):
             res = x0
             fsolve_info = 'taken_guess'
@@ -616,7 +663,7 @@ def find_closed_orbit(tracker, particle_co_guess=None, particle_ref=None,
             break
 
         (res, infodict, ier, mesg
-            ) = fsolve(lambda p: _error_for_co(p, particle_co_guess, tracker, delta_zeta, delta0),
+            ) = fsolve(lambda p: _error_for_co(p, particle_co_guess, tracker, delta_zeta, delta0, zeta0),
                 x0=x0,
                 full_output=True,
                 **co_search_settings)
@@ -648,6 +695,7 @@ def _one_turn_map(p, particle_ref, tracker, delta_zeta):
     part.py = p[3]
     part.zeta = p[4] + delta_zeta
     part.delta = p[5]
+    part.at_turn = AT_TURN_FOR_TWISS
 
     tracker.track(part)
     p_res = np.array([
@@ -659,10 +707,10 @@ def _one_turn_map(p, particle_ref, tracker, delta_zeta):
            part._xobject.delta[0]])
     return p_res
 
-def _error_for_co_search_6d(p, particle_co_guess, tracker, delta_zeta, delta0):
+def _error_for_co_search_6d(p, particle_co_guess, tracker, delta_zeta, delta0, zeta0):
     return p - _one_turn_map(p, particle_co_guess, tracker, delta_zeta)
 
-def _error_for_co_search_4d(p, particle_co_guess, tracker, delta_zeta, delta0):
+def _error_for_co_search_4d_delta0(p, particle_co_guess, tracker, delta_zeta, delta0, zeta0):
     one_turn_res = _one_turn_map(p, particle_co_guess, tracker, delta_zeta)
     return np.array([
         p[0] - one_turn_res[0],
@@ -672,6 +720,26 @@ def _error_for_co_search_4d(p, particle_co_guess, tracker, delta_zeta, delta0):
         0,
         p[5] - delta0])
 
+def _error_for_co_search_4d_zeta0(p, particle_co_guess, tracker, delta_zeta, delta0, zeta0):
+    one_turn_res = _one_turn_map(p, particle_co_guess, tracker, delta_zeta)
+    return np.array([
+        p[0] - one_turn_res[0],
+        p[1] - one_turn_res[1],
+        p[2] - one_turn_res[2],
+        p[3] - one_turn_res[3],
+        p[4] - zeta0,
+        0])
+
+def _error_for_co_search_4d_delta0_zeta0(p, particle_co_guess, tracker, delta_zeta, delta0, zeta0):
+    one_turn_res = _one_turn_map(p, particle_co_guess, tracker, delta_zeta)
+    return np.array([
+        p[0] - one_turn_res[0],
+        p[1] - one_turn_res[1],
+        p[2] - one_turn_res[2],
+        p[3] - one_turn_res[3],
+        p[4] - zeta0,
+        p[5] - delta0])
+
 def compute_one_turn_matrix_finite_differences(
         tracker, particle_on_co,
         steps_r_matrix=None):
@@ -714,6 +782,8 @@ def compute_one_turn_matrix_finite_differences(
         part_temp.s[:] = particle_on_co._xobject.s[0]
         part_temp.at_element[:] = particle_on_co._xobject.at_element[0]
 
+    part_temp.at_turn = AT_TURN_FOR_TWISS
+
     if particle_on_co._xobject.at_element[0]>0:
         i_start = particle_on_co._xobject.at_element[0]
         tracker.track(part_temp, ele_start=i_start)
@@ -766,37 +836,38 @@ def _build_auxiliary_tracker_with_extra_markers(tracker, at_s, marker_prefix,
         for nn, mm, ss in zip(names_inserted_markers, markers, at_s):
             auxline.insert_element(element=mm, name=nn, at_s=ss)
     elif algorithm == 'regen_all_drifts':
-        s_elems = auxline.get_s_elements()
-        s_keep = []
-        enames_keep = []
-        for ss, nn in zip(s_elems, auxline.element_names):
-            if not (_behaves_like_drift(auxline[nn]) and np.abs(auxline[nn].length)>0):
-                s_keep.append(ss)
-                enames_keep.append(nn)
-                assert not xt.line._is_thick(auxline[nn]) or auxline[nn].length == 0
+        raise ValueError # This algorithm is not enabled since it is not fully tested
+        # s_elems = auxline.get_s_elements()
+        # s_keep = []
+        # enames_keep = []
+        # for ss, nn in zip(s_elems, auxline.element_names):
+        #     if not (_behaves_like_drift(auxline[nn]) and np.abs(auxline[nn].length)>0):
+        #         s_keep.append(ss)
+        #         enames_keep.append(nn)
+        #         assert not xt.line._is_thick(auxline[nn]) or auxline[nn].length == 0
 
-        s_keep.extend(list(at_s))
-        enames_keep.extend(names_inserted_markers)
+        # s_keep.extend(list(at_s))
+        # enames_keep.extend(names_inserted_markers)
 
-        ind_sorted = np.argsort(s_keep)
-        s_keep = np.take(s_keep, ind_sorted)
-        enames_keep = np.take(enames_keep, ind_sorted)
+        # ind_sorted = np.argsort(s_keep)
+        # s_keep = np.take(s_keep, ind_sorted)
+        # enames_keep = np.take(enames_keep, ind_sorted)
 
-        i_new_drift = 0
-        new_enames = []
-        new_ele_dict = auxline.element_dict.copy()
-        new_ele_dict.update({nn: ee for nn, ee in zip(names_inserted_markers, markers)})
-        s_curr = 0
-        for ss, nn in zip(s_keep, enames_keep):
-            if ss > s_curr + 1e-6:
-                new_drift = xt.Drift(length=ss-s_curr)
-                new_dname = f'_auxrift_{i_new_drift}'
-                new_ele_dict[new_dname] = new_drift
-                new_enames.append(new_dname)
-                i_new_drift += 1
-                s_curr = ss
-            new_enames.append(nn)
-        auxline = xt.Line(elements=new_ele_dict, element_names=new_enames)
+        # i_new_drift = 0
+        # new_enames = []
+        # new_ele_dict = auxline.element_dict.copy()
+        # new_ele_dict.update({nn: ee for nn, ee in zip(names_inserted_markers, markers)})
+        # s_curr = 0
+        # for ss, nn in zip(s_keep, enames_keep):
+        #     if ss > s_curr + 1e-6:
+        #         new_drift = xt.Drift(length=ss-s_curr)
+        #         new_dname = f'_auxrift_{i_new_drift}'
+        #         new_ele_dict[new_dname] = new_drift
+        #         new_enames.append(new_dname)
+        #         i_new_drift += 1
+        #         s_curr = ss
+        #     new_enames.append(nn)
+        # auxline = xt.Line(elements=new_ele_dict, element_names=new_enames)
 
     auxtracker = xt.Tracker(
         _buffer=tracker._buffer,
@@ -812,6 +883,9 @@ def _build_auxiliary_tracker_with_extra_markers(tracker, at_s, marker_prefix,
     )
     auxtracker.matrix_responsiveness_tol = tracker.matrix_responsiveness_tol
     auxtracker.matrix_stability_tol = tracker.matrix_stability_tol
+    auxtracker.config = tracker.config.copy()
+    auxtracker._radiation_model = tracker._radiation_model
+    auxtracker._beamstrahlung_model = tracker._beamstrahlung_model
 
     return auxtracker, names_inserted_markers
 
@@ -912,6 +986,79 @@ class TwissTable(Table):
 
         return res
 
+    def get_normalized_coordinates(self, particles, nemitt_x=None, nemitt_y=None,
+                                   _force_at_element=None):
+
+        # TODO: check consistency of gamma0
+
+        if nemitt_x is None:
+            gemitt_x = 1
+        else:
+            gemitt_x = (nemitt_x / particles._xobject.beta0[0]
+                        / particles._xobject.gamma0[0])
+
+        if nemitt_y is None:
+            gemitt_y = 1
+        else:
+            gemitt_y = (nemitt_y / particles._xobject.beta0[0]
+                        / particles._xobject.gamma0[0])
+
+
+        ctx2np = particles._context.nparray_from_context_array
+        at_element_particles = ctx2np(particles.at_element)
+
+        part_id = ctx2np(particles.particle_id).copy()
+        at_element = part_id.copy() * 0 + xp.particles.LAST_INVALID_STATE
+        x_norm = ctx2np(particles.x).copy() * 0 + xp.particles.LAST_INVALID_STATE
+        px_norm = x_norm.copy()
+        y_norm = x_norm.copy()
+        py_norm = x_norm.copy()
+        zeta_norm = x_norm.copy()
+        pzeta_norm = x_norm.copy()
+
+        at_element_no_rep = list(set(
+            at_element_particles[part_id > xp.particles.LAST_INVALID_STATE]))
+
+        for at_ele in at_element_no_rep:
+
+            if _force_at_element is not None:
+                at_ele = _force_at_element
+
+            W = self.W_matrix[at_ele]
+            W_inv = np.linalg.inv(W)
+
+            mask_at_ele = at_element_particles == at_ele
+
+            if _force_at_element is not None:
+                mask_at_ele = ctx2np(particles.state) > xp.particles.LAST_INVALID_STATE
+
+            n_at_ele = np.sum(mask_at_ele)
+
+            # Coordinates wrt to the closed orbit
+            XX = np.zeros(shape=(6, n_at_ele), dtype=np.float64)
+            XX[0, :] = ctx2np(particles.x)[mask_at_ele] - self.x[at_ele]
+            XX[1, :] = ctx2np(particles.px)[mask_at_ele] - self.px[at_ele]
+            XX[2, :] = ctx2np(particles.y)[mask_at_ele] - self.y[at_ele]
+            XX[3, :] = ctx2np(particles.py)[mask_at_ele] - self.py[at_ele]
+            XX[4, :] = ctx2np(particles.zeta)[mask_at_ele] - self.zeta[at_ele]
+            XX[5, :] = ((ctx2np(particles.ptau)[mask_at_ele] - self.ptau[at_ele])
+                        / particles._xobject.beta0[0])
+
+            XX_norm = np.dot(W_inv, XX)
+
+            x_norm[mask_at_ele] = XX_norm[0, :] / np.sqrt(gemitt_x)
+            px_norm[mask_at_ele] = XX_norm[1, :] / np.sqrt(gemitt_x)
+            y_norm[mask_at_ele] = XX_norm[2, :] / np.sqrt(gemitt_y)
+            py_norm[mask_at_ele] = XX_norm[3, :] / np.sqrt(gemitt_y)
+            zeta_norm[mask_at_ele] = XX_norm[4, :]
+            pzeta_norm[mask_at_ele] = XX_norm[5, :]
+            at_element[mask_at_ele] = at_ele
+
+        return Table({'particle_id': part_id, 'at_element': at_element,
+                      'x_norm': x_norm, 'px_norm': px_norm, 'y_norm': y_norm,
+                      'py_norm': py_norm, 'zeta_norm': zeta_norm,
+                      'pzeta_norm': pzeta_norm})
+
     def reverse(self):
 
         assert self.values_at == 'entry', 'Not yet implemented for exit'
@@ -1005,35 +1152,6 @@ class TwissTable(Table):
                     else:
                         self[kk] = [vv[ii] for ii in indx_twiss]
 
-def _error_for_match(knob_values, vary, targets, tracker, tw_kwargs):
-    for kk, vv in zip(vary, knob_values):
-        tracker.vars[kk] = vv
-    tw = tracker.twiss(**tw_kwargs)
-    res = []
-    for tt in targets:
-        if isinstance(tt[0], str):
-            res.append(tw[tt[0]] - tt[1])
-        else:
-            res.append(tt[0](tw) - tt[1])
-    return np.array(res)
-
-def match_tracker(tracker, vary, targets, **kwargs):
-    _err = partial(_error_for_match, vary=vary, targets=targets,
-                   tracker=tracker, tw_kwargs=kwargs)
-    x0 = [tracker.vars[vv]._value for vv in vary]
-    try:
-        (res, infodict, ier, mesg) = fsolve(_err, x0=x0.copy(), full_output=True)
-        if ier != 1:
-            raise RuntimeError("fsolve failed: %s" % mesg)
-        for kk, vv in zip(vary, res):
-            tracker.vars[kk] = vv
-        fsolve_info = {
-                'res': res, 'info': infodict, 'ier': ier, 'mesg': mesg}
-    except Exception as err:
-        for ii, vv in enumerate(vary):
-            tracker.vars[vv] = x0[ii]
-        raise err
-    return fsolve_info
 
 def _renormalize_eigenvectors(Ws):
     # Re normalize eigenvectors
@@ -1080,7 +1198,10 @@ def _renormalize_eigenvectors(Ws):
 
 def _extract_twiss_parameters_with_inverse(Ws):
 
-    BB = np.zeros(shape=(3, Ws.shape[0], 6, 6), dtype=np.float64)
+    # From E. Forest, "From tracking code to analysis", Sec 4.1.2 or better
+    # https://iopscience.iop.org/article/10.1088/1748-0221/7/07/P07012
+
+    EE = np.zeros(shape=(3, Ws.shape[0], 6, 6), dtype=np.float64)
 
     for ii in range(3):
         Iii = np.zeros(shape=(6, 6))
@@ -1090,14 +1211,17 @@ def _extract_twiss_parameters_with_inverse(Ws):
 
         Ws_inv = np.linalg.inv(Ws)
 
-        BB[ii, :, :, :] = Ws @ Sii @ Ws_inv
+        EE[ii, :, :, :] = - Ws @ Sii @ Ws_inv @ lnf.S
 
-    betx = BB[0, :, 0, 1]
-    bety = BB[1, :, 2, 3]
-    alfx = BB[0, :, 0, 0]
-    alfy = BB[1, :, 2, 2]
-    gamx = -BB[0, :, 1, 0]
-    gamy = -BB[1, :, 3, 2]
+    betx = EE[0, :, 0, 0]
+    bety = EE[1, :, 2, 2]
+    alfx = -EE[0, :, 0, 1]
+    alfy = -EE[1, :, 2, 3]
+    gamx = EE[0, :, 1, 1]
+    gamy = EE[1, :, 3, 3]
+
+    bety1 = np.abs(EE[0, :, 2, 2])
+    betx2 = np.abs(EE[1, :, 0, 0])
 
     sign_x = np.sign(betx)
     sign_y = np.sign(bety)
@@ -1108,4 +1232,4 @@ def _extract_twiss_parameters_with_inverse(Ws):
     alfy *= sign_y
     gamy *= sign_y
 
-    return betx, alfx, gamx, bety, alfy, gamy
\ No newline at end of file
+    return betx, alfx, gamx, bety, alfy, gamy, bety1, betx2
